#include "mainwindow.h"
#include "./ui_mainwindow.h"

#include <iostream>

#include <QClipboard>
#include <QtConcurrent>
#include <QErrorMessage>
#include <QFile>
#include <QFileDialog>
#include <QFont>
#include <QFontMetrics>
#include <QFuture>
#include <QStyle>

#include "aboutwindow.h"
#include "fossgis.h"
#include "openrouteservice.h"


// MainWindow — The primary window used by the application.
MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent),
      ui(new Ui::MainWindow)
{
    addresser = nullptr;
    currentRouter = nullptr;
    calculating = false;

    orsMapData = nullptr;
    osrmMapData = nullptr;
    rkMapDataDriving = nullptr;
    rkMapDataWalking = nullptr;

    setupUi();
}


MainWindow::~MainWindow()
{
    delete ui;
}


// Create the appropriate Router object to handle calculation of distances between
// points, then have it start calculations; several signals from the Router object
// will be handled, most importantly `distancesCalculated`.
// distancesCalculated => distancesCalculated
// calculationError => calculationError
// calculationProgress => calculationProgress
void
MainWindow::calculateDistances()
{
    vector_2d_double points = coordListTo2dVector();
    QString engineName = currentEngine();
    Router* router = nullptr;

    calculating = true;
    validateOverall();

    // Save the currently-listed addresses for later display.
    previousAddresses.clear();
    for (int i = 0; i < ui->addressList->count(); i++)
	previousAddresses << ui->addressList->item(i)->text();

    if (points.size() == 1) {
	QErrorMessage* error = new QErrorMessage();
	error->showMessage("Pro výpočet sazeb prosím zadejte alespoň dvě lokace.");
	return;
    }

    // Select the appropriate Router, and initialize it.
    if (engineName == "routingKit") {
	loadRkMapData();
	if (travelByCar() && rkMapDataDriving)
	    router = new RoutingKitRouter<RKDrivingGraph>(rkMapDataDriving);
	else if (rkMapDataWalking)
	    router = new RoutingKitRouter<RKWalkingGraph>(rkMapDataWalking);
    }
    else if (engineName == "osmrNet") {
	loadOsrmMapData();
	if (osrmMapData != nullptr) {
	    osrmMapData->byCar = travelByCar();
	    router = new FOSSGISRouter(osrmMapData);
	}
    }
    else if (engineName == "openRouteService") {
	loadOrsMapData();
	if (orsMapData != nullptr) {
	    orsMapData->byCar = travelByCar();
	    router = new OpenRouteServiceRouter(orsMapData);
	}
    }

    // If a Router was initialized, kick off the distance calculation.
    if (router != nullptr) {
	delete currentRouter;
	currentRouter = router;

	validateOverall();

	connect(router, &Router::distancesCalculated,
		this, &MainWindow::distancesCalculated);
	connect(router, &Router::calculationProgress,
		this, &MainWindow::calculationProgress);
	connect(router, &Router::calculationError,
		this, &MainWindow::calculationError);

	router->calculateDistances(points);
    }
    else
	calculating = false;
}


// Run `calculateDistances` in a seperate thread, so the UI doesn't become unresponsive.
void
MainWindow::beginCalculateDistances()
{
    if (calculating == false)
	QFuture future = QtConcurrent::run(&MainWindow::calculateDistances, this);
}


// Display any errors signaled from the calculating Router.
// Handles signal Router::calculationError.
void
MainWindow::calculationError(QString text)
{
    std::cerr << text.toStdString() << std::endl;

    QErrorMessage* error = new QErrorMessage();
    QString errorMsg = text;
    if (errorMsg.isEmpty())
	errorMsg = "Výpočet se nezdařil.\r\n";
    error->showMessage(errorMsg);

    ui->progressBar->setValue(0);
    ui->progressBar->setFormat("");
    calculating = false;
}


// Display current progress of calculation; either as a flat percent, or as a number
// to add to the current percent-progress.
// Handles signal Router::calculationProgress.
void
MainWindow::calculationProgress(float percent, float percentIncrement, QString text)
{
    std::cerr << text.toStdString() << std::endl;

    if (percentIncrement > 0)
	ui->progressBar->setValue(ui->progressBar->value() + percentIncrement);
    else
	ui->progressBar->setValue(percent);

    ui->progressBar->setFormat(text);

}


// When distances are calculated, tweak the UI and display them.
// Handles signal Router::distancesCalculated.
void
MainWindow::distancesCalculated(vector_2d_double points, vector_2d_int distances)
{
    previousPoints = points;
    previousDistances = distances;

    ui->progressBar->setValue(100);
    ui->progressBar->setFormat("");
    calculating = false;

    displayDistances(points, distances);
    ui->copyCsvButton->setEnabled(true);

    validateOverall();
}


// When an error has occured during loading of map data, tell the user.
// Handles signal MapData::mapDataError.
void
MainWindow::mapDataError(QString text)
{
    calculationError(text);
    validateOverall();
}


// When progress has been made in the loading of map data, tell the user.
// Handles signal MapData::mapDataProgress.
void
MainWindow::mapDataProgress(float percent, float percentIncrement, QString text)
{
    calculationProgress(percent, percentIncrement, text);
}


// When map data has finished being parsed, mark it as such.
// Handles signal MapData::mapDataProgress.
void
MainWindow::mapDataFinished()
{
    ui->progressBar->setValue(100);
    ui->progressBar->setFormat("");

    validateOverall();
}


// Loads RoutingKit MapData for the current file and routing type (walking/driving).
void
MainWindow::loadRkMapData()
{
    std::string path = ui->rkMapfileInput->text().toStdString();

    if (rkMapDataDriving == nullptr && travelByCar()) {
	rkMapDataDriving = new RoutingKitMapData<RKDrivingGraph>(true, path);
	connectMapData(rkMapDataDriving);
	if (!rkMapDataDriving->loadMap()) {
	    delete rkMapDataDriving;
	    rkMapDataDriving = nullptr;
	}
    }
    else if (rkMapDataWalking == nullptr && !travelByCar()) {
	rkMapDataWalking = new RoutingKitMapData<RKWalkingGraph>(false, path);
	connectMapData(rkMapDataWalking);
	if (!rkMapDataWalking->loadMap()) {
	    delete rkMapDataWalking;
	    rkMapDataWalking = nullptr;
	}
    }
}


// “Loads” OpenRouteService MapData for the current API-key and server.
void
MainWindow::loadOrsMapData()
{
    QString api_key = ui->orsUseDefaultApiInput->isChecked() ? ORS_API_KEY : ui->orsApiKeyInput->text();
    QString server = ui->orsServerInput->text();
    if (orsMapData == nullptr && !api_key.isEmpty() && !server.isEmpty()) {
	orsMapData = new OpenRouteServiceMapData(server, api_key);
	if (!orsMapData->loadMap()) {
	    delete orsMapData;
	    orsMapData = nullptr;
	}
    }
}


// “Loads” FOSSGIS/OSRM MapData for the current server.
void
MainWindow::loadOsrmMapData()
{
    if (osrmMapData == nullptr && ui->osmrServerInput->text().isEmpty() == false) {
	osrmMapData = new FOSSGISMapData(ui->osmrServerInput->text());
	if (!osrmMapData->loadMap()) {
	    delete osrmMapData;
	    osrmMapData = nullptr;
	}
    }
}


// Connect a MapData object's signals to the MainWindow's signal-handlers.
void
MainWindow::connectMapData(MapData* mapdata)
{
    connect(mapdata, &MapData::mapDataError,
	    this, &MainWindow::mapDataError);
    connect(mapdata, &MapData::mapDataProgress,
	    this, &MainWindow::mapDataProgress);
    connect(mapdata, &MapData::mapDataFinished,
	    this, &MainWindow::mapDataFinished);
}


// Print distances as CSV to stdout, and populate the window's table/CSV text-box with them.
void
MainWindow::displayDistances(vector_2d_double points, vector_2d_int distances)
{
    if (points.size() == distances.size() && points.size() >= 2) {
    	print_point_distances_csv(points, distances);
	populateTableWithDistances(points, distances);
	populateTextBrowserWithDistances(points, distances);
    }
}


// Open the “About” window, if not already open.
// Handles signal QAction::toggled (actionAbout).
void
MainWindow::openAboutWindow()
{
    AboutWindow* about = new AboutWindow(this);
    about->open();
}


// Copies the CSV data to clipboard.
// Handles signals QAction::toggled (actionCopyCSV), QPushButton::clicked (copyCsvButton).
void
MainWindow::copyCSVData()
{
    QGuiApplication::clipboard()->setText(ui->distanceText->toPlainText());
}


// Clears all locations from the location-list.
// Handles signal QAction::toggled (actionClearLocations)
void
MainWindow::clearLocations()
{
    ui->coordinateList->clear();
    ui->addressList->clear();
}


// If CSV/table formatting has been changed, re-populate the table and text ouput.
// Handles signal QAction::toggled (actionCoordsInCS, actionShowNames)
void
MainWindow::repopulateTableAndCSV(bool checked)
{
    if (!calculating) {
	populateTableWithDistances(previousPoints, previousDistances);
	populateTextBrowserWithDistances(previousPoints, previousDistances);
    }
}


// Add a the string currently in the "address input" textbox to the address list, if valid.
// Handles signals QLineEdit::returnPressed, QPushButton::clicked (addressInput, addressAddButton).
void
MainWindow::addLocationString()
{
    QLineEdit* input = ui->addressInput;
    if (is_coord_string(input->text())) {
	addToLocationsList(coord_string_to_vector(input->text()), "");
	input->clear();
    }
    else if (ui->orsTranslateToCoordsInput->isChecked()) {
	ui->addressAddButton->setDisabled(true);
	input->setDisabled(true);

	resetAddresser();
	addresser->locateAddress(input->text());
    }
}


// Inform the user, if a coordinate couldn't be added by address.
// Handles singal Addresser:locationError (addresser).
void
MainWindow::addLocationAddressError(QString message)
{
    QErrorMessage* error = new QErrorMessage();
    QString errorMsg = message;
    if (errorMsg.isEmpty())
	errorMsg = "Nepodařilo se převést adresu na souřadnice.\r\n";
    error->showMessage(errorMsg);

   ui->addressInput->setEnabled(true);
   ui->addressAddButton->setEnabled(true);
   validateOverall();
}


// Re-enable address textbox if the Addressor's given up on translating the address.
void
MainWindow::addLocationAddressCancelled()
{
    ui->addressInput->setEnabled(true);
    ui->addressAddButton->setEnabled(true);
    validateOverall();
}


// Add an Addresser's calculated coordinate (from address string) to the coord-list.
// Handles signal Addresser::locationFound (addresser).
void
MainWindow::addLocationCoordinates(std::vector<double> point, QString name)
{
    addToLocationsList(point, name);
    ui->addressInput->clear();
    ui->addressInput->setEnabled(true);
}


// Removes the currently selected location from the lists.
// Handles signal QPushButton:clicked (addressDelButton).
void
MainWindow::removeSelectedLocation()
{
    int selectionA = ui->coordinateList->currentRow();
    int selectionB = ui->addressList->currentRow();
    int selection = selectionA >= 0 ? selectionA : selectionB;
    if (selection >= 0) {
	ui->coordinateList->takeItem(selection);
	ui->addressList->takeItem(selection);
    }
}


// Open a file-dialog window to select a PBF file for the RoutingKitRouter.
// Handles QPushButton::clicked (rkMapfileExploreButton).
void
MainWindow::rkBrowseMapFile()
{
    QString mapPath =
	QFileDialog::getOpenFileName(this, "Otevřít soubor s mapou",
				     ".", "Mapy OSM (*.pbf)");
    ui->rkMapfileInput->setText(mapPath);

    delete rkMapDataDriving;
    delete rkMapDataWalking;
    rkMapDataDriving = nullptr;
    rkMapDataWalking = nullptr;

    validateOverall();
}


// Make sure to reset current RoutingKit data if the selected file changes.
// Handles signal QLineEdit::textChanged (rkMapfileInput).
void
MainWindow::rkFileChanged(const QString& text)
{
    delete rkMapDataDriving;
    delete rkMapDataWalking;
    rkMapDataDriving = nullptr;
    rkMapDataWalking = nullptr;

    validateOverall();
}


// If the ORS server or API key have changed, make sure to reset the current ORS data,
// so this data can be processed before distance calculation.
// Handles signal QLineEdit::textChanged (orsServerInput, orsApikeyInput).
void
MainWindow::orsApiOrServerChanged(const QString& text)
{
    delete orsMapData;
    orsMapData = nullptr;
    validateOverall();
}


// Update the state of API-textbox input, depending on the "default API" checkbox.
// Handles signal QCheckbox::stateChanged (orsUseDefaultApiInput).
void
MainWindow::orsApiKeyToggled(int state)
{
    ui->orsApiKeyInput->setDisabled(state);
    if (state)
	ui->orsApiKeyInput->setPlaceholderText("Bude použit výchozí, přednastavený klíč.");
    else
	ui->orsApiKeyInput->setPlaceholderText("Vložte váš klíč API…");

    delete orsMapData;
    orsMapData = nullptr;
    validateOverall();
}


// Update the placeholder text of address textbox, based on state of address translation.
// Handles signal QCheckbox::stateChanged (orsTranslateToCoordsInput).
void
MainWindow::orsGeocodingToggled(int state)
{
    if (state)
	ui->addressInput->setPlaceholderText("Zadejte adresu či souřadnice…");
    else
	ui->addressInput->setPlaceholderText("Zadejte souřadnice…");

    validateOverall();
}


// Make sure to reset OSRM data if the server is changed by the user.
// Handles signal QLineEdit::textChanged (osmrServerInput).
void
MainWindow::osrmServerChanged(const QString& text)
{
    delete osrmMapData;
    osrmMapData = nullptr;
    validateOverall();
}


// Disable the "add address" button if given string is invalid address.
// Handles QLineEdit::textChanged (addressInput).
void
MainWindow::locationAddressChanged(const QString& text)
{
    if (ui->orsTranslateToCoordsInput->isChecked())
	ui->addressAddButton->setDisabled(text.isEmpty());
    else
	ui->addressAddButton->setEnabled(is_coord_string(text));
    validateOverall();
}


// Update UI disable/enable state, when the engine tab changes.
// Handles QTabWidget::currentChanged (enginesTabWidget).
void
MainWindow::engineSelected(int tabIndex)
{
    validateOverall();
}


// Enable/disable the location-removal button, based on whether or not a location is selected
// in the lists.
// Handles signal QListWidget::currentRowChanged (addressList, coordinateList).
void
MainWindow::locationSelected(int row)
{
    ui->addressDelButton->setEnabled(row >= 0);
    ui->addressList->setCurrentRow(row);
    ui->coordinateList->setCurrentRow(row);
}


// Validates input generally, making sure that the current settings (according to tab) are suitable
// for starting route calculation. This is called by most every input-validation method in this class.
bool
MainWindow::validateOverall()
{
    QString engineName = currentEngine();
    bool valid = true;

    if (calculating) {
	ui->travelTypeCombo->setEnabled(false);
	valid = false;
    }
    else
	ui->travelTypeCombo->setEnabled(true);

    if (engineName == "routingKit") {
	if (QFile(ui->rkMapfileInput->text()).exists() == false)
	    valid = false;
    }
    else if (engineName == "osmrNet" || engineName == "openRouteService") {
	QString urlStr = (engineName == "osmrNet") ? ui->osmrServerInput->text() : ui->orsServerInput->text();
	QUrl url(urlStr);
	if (url.isValid() == false || (url.scheme() != "http" && url.scheme() != "https"))
	    valid = false;

	if (engineName == "openRouteService"
	    && ui->orsUseDefaultApiInput->isChecked() == false && ui->orsApiKeyInput->text().isEmpty())
	    valid = false;
    }

    if (ui->coordinateList->count() == 0)
	valid = false;

    ui->calculateButton->setEnabled(valid);
    return valid;
}


// Return the name of the currently-selected routing engine (tab).
QString
MainWindow::currentEngine()
{
    return ui->enginesTabWidget->currentWidget()->objectName();
}


// There are two options for route-calculation: By car, or by foot.
// This returns true if the current option is car; implicitly, false means the current option is foot.
bool
MainWindow::travelByCar()
{
    return (ui->travelTypeCombo->currentIndex() == 0);
}


// Add a point (with associated address-label) to the point and label lists.
void
MainWindow::addToLocationsList(std::vector<double> point, QString name)
{
    QListWidget* coordList = ui->coordinateList;
    QListWidget* addrList = ui->addressList;

    QString pointStr = coord_vector_to_string(point);
    for (int i = 0; i < coordList->count(); i++) {
	if (coordList->item(i)->text() == pointStr)
	    return;
    }

    ui->coordinateList->addItem(pointStr);
    ui->addressList->addItem(name);
}


// Populates the distance-table with the given data.
void
MainWindow::populateTableWithDistances(vector_2d_double points, vector_2d_int distances)
{
    QTableWidget* table = ui->distanceTable;
    table->clear();
    table->setRowCount(points.size());
    table->setColumnCount(points.size());

    bool do_display_addresses = ui->actionShowNames->isChecked();

    for (int i = 0; i < points.size(); i++) {
	// Populate table row/column headers.
	QString header = (do_display_addresses && previousAddresses[i].isEmpty() == false)
	    ? previousAddresses[i]
	    : coord_vector_to_string(points[i]);
	table->setVerticalHeaderItem(i, new QTableWidgetItem(header));
	table->setHorizontalHeaderItem(i, new QTableWidgetItem(header));

	// Populate the actual table data
	for (int j = 0; j < points.size(); j++) {
	    QString distanceStr = QString().setNum(distances[i][j]).append(" m");
	    table->setItem(i, j, new QTableWidgetItem(distanceStr));
	}
    }
}


// Populates the CSV textbox with this given distance data.
void
MainWindow::populateTextBrowserWithDistances(vector_2d_double points, vector_2d_int distances)
{
    int pointc = points.size();
    QTextBrowser* text = ui->distanceText;
    text->clear();

    bool do_display_coordinates = ui->actionCoordsInCSV->isChecked();
    bool do_display_addresses = ui->actionShowNames->isChecked();

    // Populate the column headers.
    if (do_display_coordinates) {
	text->insertPlainText(","); // First header ought to be empty.

	for (int i = 0; i < pointc; i++) {
	    QString header = do_display_addresses
		? coord_vector_w_label_to_string(points[i], previousAddresses[i])
		: coord_vector_to_colon_delimited_string(points[i]);

	    text->insertPlainText(header);
	    if (i < pointc - 1 && i < pointc)
		text->insertPlainText(",");
	}
	text->insertPlainText("\r\n");
    }

    for (int i = 0; i < pointc; i++) {
	// Populate the row's header.
	if (do_display_coordinates) {
	    QString header = do_display_addresses
		? coord_vector_w_label_to_string(points[i], previousAddresses[i])
		: coord_vector_to_colon_delimited_string(points[i]);

	    text->insertPlainText(header);
	    text->insertPlainText(",");
	}

	// Populate the row's data.
	for (int j = 0; j < pointc; j++) {
	    QString distanceStr = QString().setNum(distances[i][j]);
	    text->insertPlainText(distanceStr);
	    if (j < pointc - 1 && i < pointc)
		text->insertPlainText(",");
	    else
		text->insertPlainText("\r\n");
	}
    }
}


// Compose all elements of the GUI's address list to a 2D vector of coordinates.
vector_2d_double
MainWindow::coordListTo2dVector()
{
    vector_2d_double coordVector;
    QListWidget* coordList = ui->coordinateList;

    for (int i = 0; i < coordList->count(); i++) {
	QListWidgetItem* item = coordList->item(i);
	coordVector.push_back(coord_string_to_vector(item->text()));
    }
    return coordVector;
}


// Create a new addresser for address-to-coordinate translations.
void
MainWindow::resetAddresser()
{
    loadOrsMapData();
    if (orsMapData != nullptr) {
	addresser = new OpenRouteServiceAddresser(orsMapData);

	connect(addresser, &Addresser::locationFound,
		this, &MainWindow::addLocationCoordinates);
	connect(addresser, &Addresser::locationError,
		this, &MainWindow::addLocationAddressError);
	connect(addresser, &Addresser::locationCancelled,
		this, &MainWindow::addLocationAddressCancelled);
    }
}


// Initialize the UI; set default values; connect signals, etc.
void
MainWindow::setupUi()
{
    ui->setupUi(this);

    ui->orsApiKeyInput->setEnabled(false);
    ui->orsUseDefaultApiInput->setChecked(true);
    ui->orsTranslateToCoordsInput->setChecked(true);
    ui->copyCsvButton->setEnabled(false);

    // The “File”-menu.
    connect(ui->actionAbout, &QAction::triggered,
	    this, &MainWindow::openAboutWindow);
    connect(ui->actionQuit, &QAction::triggered,
	    this, &QWidget::close);

    // The “Edit”-menu.
    connect(ui->actionCopyCSV, &QAction::triggered,
	    this, &MainWindow::copyCSVData);
    connect(ui->actionClearLocations, &QAction::triggered,
	    this, &MainWindow::clearLocations);

    // The “View”-menu
    connect(ui->actionShowNames, &QAction::toggled,
	    this, &MainWindow::repopulateTableAndCSV);
    connect(ui->actionCoordsInCSV, &QAction::toggled,
	    this, &MainWindow::repopulateTableAndCSV);

    // Copy CSV data when button pushed.
    connect(ui->copyCsvButton, &QPushButton::clicked,
	    this, &MainWindow::copyCSVData);

    // Calculate distances when button pushed.
    connect(ui->calculateButton, &QPushButton::clicked,
	    this, &MainWindow::beginCalculateDistances);

    // Disable the “add address” button when input is invalid.
    connect(ui->addressInput, &QLineEdit::textChanged,
	    this, &MainWindow::locationAddressChanged);

    // Add an input address to the address list.
    connect(ui->addressInput, &QLineEdit::returnPressed,
	    this, &MainWindow::addLocationString);
    connect(ui->addressAddButton, &QPushButton::clicked,
	    this, &MainWindow::addLocationString);

    // Enable deletion of items from the address list
    connect(ui->addressList, &QListWidget::currentRowChanged,
	    this, &MainWindow::locationSelected);
    connect(ui->coordinateList, &QListWidget::currentRowChanged,
	    this, &MainWindow::locationSelected);
    connect(ui->addressDelButton, &QPushButton::clicked,
	    this, &MainWindow::removeSelectedLocation);

    // Ensure the URL for a server is valid.
    connect(ui->osmrServerInput, &QLineEdit::textChanged,
	    this, &MainWindow::osrmServerChanged);
    connect(ui->orsServerInput, &QLineEdit::textChanged,
	    this, &MainWindow::orsApiOrServerChanged);

    // Worry about ORS's API-key field.
    connect(ui->orsApiKeyInput, &QLineEdit::textChanged,
	    this, &MainWindow::orsApiOrServerChanged);
    connect(ui->orsUseDefaultApiInput, &QCheckBox::stateChanged,
	    this, &MainWindow::orsApiKeyToggled);

    // … and now about its support for address => coordinate translation…
    connect(ui->orsTranslateToCoordsInput, &QCheckBox::stateChanged,
	    this, &MainWindow::orsGeocodingToggled);

    // Ensure the selected file for RoutingKit exists and is of the appropriate type.
    connect(ui->rkMapfileExploreButton, &QPushButton::clicked,
	    this, &MainWindow::rkBrowseMapFile);
    connect(ui->rkMapfileInput, &QLineEdit::textChanged,
	    this, &MainWindow::rkFileChanged);

    // Ensure the selected engine can be used immediately
    connect(ui->enginesTabWidget, &QTabWidget::currentChanged,
	    this, &MainWindow::engineSelected);


    // Set and/or tweak button icons.
    set_button_qt_icon(ui->rkMapfileExploreButton, QStyle::SP_DirOpenIcon);

    use_button_icon_or_text(ui->copyCsvButton);
    use_button_icon_or_text(ui->addressAddButton);
    use_button_icon_or_text(ui->addressDelButton);

    orsGeocodingToggled(true);
    validateOverall();
}



// For buttons that should be icon or text-only: This either hides the text or disables
// the icon, depending on whether or not the button's icon is valid.
void
use_button_icon_or_text(QAbstractButton* button)
{
    if (button->icon().pixmap(20).isNull())
	button->setIconSize(QSize(0, 0));
    else {
	// Icons sometimes enlargen buttons; let's make sure they accord with the font-size.
	QFont font = button->font();
	int height = QFontMetrics(font).height();

	button->setText("");
	button->setIconSize(QSize(height, height));
    }
}


// Sets a button's icon to a built-in QT standard icon.
void
set_button_qt_icon(QAbstractButton* button, QStyle::StandardPixmap icon)
{
    button->setIcon(
	button->style()->standardIcon(icon));
}

