#include "routingkit.h"

#include <iostream>
#include <string>
#include <vector>

#include <QFile>


// Logging method passed to RoutingKit's calculations for debugging.
void
log_message(const std::string& str)
{
    std::cerr << str << std::endl;
}


template class RoutingKitMapData<RKDrivingGraph>;
template class RoutingKitMapData<RKWalkingGraph>;


// The RoutingKit class for parsing/storing map data.
// There are two sorts of interal graph datatypes used, depending on whether the
// routing is by car or by foot (hence the templating).
template<typename ROUTETYPE>
RoutingKitMapData<ROUTETYPE>::RoutingKitMapData(bool by_car, std::string map_path)
    : MapData(by_car)
{
    mapPath = map_path;
}


// Actually parse the map-data for use with a RoutingKit Router or Addresser.
template<typename ROUTETYPE>
bool
RoutingKitMapData<ROUTETYPE>::loadMap()
{
    emit mapDataProgress(0, 0, "Čtení mapových dat…");

    if (mapPath.size() == 0) {
	emit mapDataError("Nebyl zadán žádný soubor; zvolte prosím soubor s mapou ve formátu PBF.");
	return false;
    }
    else if (QFile(mapPath.c_str()).exists() == false) {
	emit mapDataError("Daný soubor neexistuje; zvolte prosím jiný.");
	return false;
    }

    if constexpr(std::is_base_of_v<ROUTETYPE, RKDrivingGraph>)
	graph = RoutingKit::simple_load_osm_car_routing_graph_from_pbf(mapPath.c_str(), &log_message);
    else if constexpr(std::is_base_of_v<ROUTETYPE, RKWalkingGraph>)
	graph = RoutingKit::simple_load_osm_pedestrian_routing_graph_from_pbf(mapPath.c_str(), &log_message);

    if (graph.node_count() == 0) {
	emit mapDataError("Čtení mapových dat selhalo. Je daný soubor ve formátu PBF?");
	return false;
    }

    emit mapDataProgress(35, 0, "Výpočet inverzních vektorů…");
    auto tail = RoutingKit::invert_inverse_vector(graph.first_out);

    emit mapDataProgress(75, 0, "Tvorba kontrakční hierarchie…");
    hierarchy = RoutingKit::ContractionHierarchy::build(graph.node_count(), tail,
							graph.head, graph.geo_distance);
    if (hierarchy.node_count() == 0 ) {
	emit mapDataError("Tvorba kontrakční hierarchie selhala. Je daný soubor ve formátu PBF?");
	return false;
    }
    emit mapDataProgress(100, 0, "Mapové údaje byly úspěšně načteny…");
    emit mapDataFinished();
    return true;
}



template class RoutingKitRouter<RKDrivingGraph>;
template class RoutingKitRouter<RKWalkingGraph>;


// The RoutingKit distance-calculation class.
template<typename ROUTETYPE>
RoutingKitRouter<ROUTETYPE>::RoutingKitRouter(RoutingKitMapData<ROUTETYPE>* mapdata)
{
    mapData = mapdata;
}


template<typename ROUTETYPE>
RoutingKitRouter<ROUTETYPE>::~RoutingKitRouter()
{
}


// Return the distance (meters) between two points on the map. This is used by
// Router::calculateDistances to find the distance matrix of several points.
template<typename ROUTETYPE>
int
RoutingKitRouter<ROUTETYPE>::pointDistance(double from_latitude, double from_longitude,
				double to_latitude, double to_longitude)
{
    RoutingKit::ContractionHierarchyQuery query(mapData->hierarchy);
    RoutingKit::GeoPositionToNode position_converter(mapData->graph.latitude,
						     mapData->graph.longitude);

    QString errorMsg("Nebyl nalezen žádný bod v blízkosti dané adresy (%1,%2).");

    unsigned from_id =
	position_converter.find_nearest_neighbor_within_radius(from_latitude, from_longitude, 2000).id;
    if (from_id == RoutingKit::invalid_id) {
	emit calculationError(errorMsg.arg(from_latitude, 0, 'g', 9).arg(from_longitude, 0, 'g', 9));
	return -2;
    }

    unsigned to_id =
	position_converter.find_nearest_neighbor_within_radius(to_latitude, to_longitude, 2000).id;
    if (to_id == RoutingKit::invalid_id) {
	emit calculationError(errorMsg.arg(to_latitude, 0, 'g', 9).arg(to_longitude, 0, 'g', 9));
	return -1;
    }

    query.reset().add_source(from_id).add_target(to_id).run();
    return query.get_distance();
}
