#include "util.h"

#include <cstdio>
#include <iomanip>
#include <iostream>

#include <QRegularExpression>
#include <QString>


// Return whether or not the given string is a valid 2D coordinate pair.
bool
is_coord_string(const QString str)
{
    QRegularExpression using_comma("^\\(?-?([0-9]+)?\\.?[0-9]+(\\s+)?,(\\s+)?-?([0-9]+)?\\.?[0-9]+\\)?$");
    QRegularExpression using_semicolon("^\\(?-?([0-9]+)?,?[0-9]+(\\s+)?;(\\s+)?-?([0-9]+)?,?[0-9]+\\)?$");

    QString sanitized = str.simplified();
    return (using_comma.match(sanitized).hasMatch()
	    || using_semicolon.match(sanitized).hasMatch());
}


// Convert a vector of 2D coordinates into a string of (a,b) format.
QString
coord_vector_to_string(std::vector<double> point)
{
    // !!! Překontrolovat; je moc hodin.
    // Precision is 9, with 6 decimal-places for accuracy < 1 meter.
    return QString("(%1,%2)").arg(point[0], 0, 'g', 9).arg(point[1], 0, 'g', 9);
}


// Convert a vector of 2D coordinates into a string, with colon as a delimiter.
QString
coord_vector_to_colon_delimited_string(std::vector<double> point)
{
    // !!! Překontrolovat; je moc hodin.
    // Precision is 9, with 6 decimal-places for accuracy < 1 meter.
    return QString("(%1;%2)").arg(point[0], 0, 'g', 9).arg(point[1], 0, 'g', 9);
}


// Convert a vector of 2D coordinates into a colon-delimited string, with an
// optional label. Ends up as: “(1.0;2.0) | label”.
QString
coord_vector_w_label_to_string(std::vector<double> point, QString label)
{
    QString pointStr = coord_vector_to_colon_delimited_string(point);
    pointStr.append(" | ");

    if (label.isEmpty() == false)
	pointStr.append(label.remove(','));

    return pointStr;
}


// Parse a coordinate pair string into a vector of doubles.
// "(1,4567; 9,8753)" or "(1.4567, 9.8573)" for example.
// If the string isn't appropriately formatted, the returned vector is empty.
std::vector<double>
coord_string_to_vector(const QString coordString)
{
    std::vector<double> coords;

    QString sanitizedStr = coordString.simplified();
    if (!is_coord_string(coordString))
	return coords;

    sanitizedStr.remove(QRegularExpression("^\\("));
    sanitizedStr.remove(QRegularExpression("\\)$"));

    QStringList parameters = sanitizedStr.split(';', Qt::SkipEmptyParts);
    if (parameters.count() == 1)
	parameters = sanitizedStr.split(',', Qt::SkipEmptyParts);
    if (parameters.count() == 1)
	return coords;

    for (auto iter = parameters.constBegin(); iter != parameters.constEnd(); ++iter) {
	QString item = QString(*iter).replace(',', '.');
	double num = item.toDouble();
	coords.push_back(num);
    }
    return coords;
}


// Print points and their associated distances from other points in
// a verbose and easy-to-read format (for debugging purposes).
void
print_point_distances_verbosely(vector_2d_double points, vector_2d_int distances,
				const char* engine)
{
    int pointc = points.size();
    for (int from_row = 0; from_row < pointc; from_row++)
	for (int to_column = 0; to_column < pointc; to_column++)
	    printf("[%s] [(%f, %f) -> (%f, %f)] %i meters.\r\n",
		   engine,
		   points[from_row][0], points[from_row][1],
		   points[to_column][0], points[to_column][1],
		   distances[from_row][to_column]);
}


// Print points and their associated distances from all other points in
// parseable CSV form.
void
print_point_distances_csv(vector_2d_double points, vector_2d_int distances)
{
    int pointc = points.size();
    for (int from_row = 0; from_row < pointc; from_row++)
	for (int to_column = 0; to_column < pointc; to_column++) {
	    printf("%i", distances[from_row][to_column]);
	    if (to_column < pointc - 1)
		printf(",");
	    else
		printf("\r\n");
	}
}


// Translate a server's generic API error message into one to be displayed to the user.
QString
generic_api_error_to_message(QString server_error)
{
    QString errorMsg("Zvolená služba odpověděla na váš požadavek chybou:\r\n");
    errorMsg.append(server_error);
    if (server_error == "Access to this API has been disallowed")
	errorMsg = "Zvolený klíč k API je neplatný – na stránkách si můžete vytvořit nový.";

    return errorMsg;
}

